#ifndef VECTOR_H
#define VECTOR_H

#include <stdint.h>
#include <cmath>

class Vector {
	private:
		double _x;
		double _y;
	public:
		enum direction {LEFT, RIGHT};

		Vector (double x = 0, double y = 0) {
			_x = x;
			_y = y;
		}

		// accessors
		double x() const {
			return _x;
		}
		double y() const {
			return _y;
		}
		
		// vector properties
		double length() {
			return sqrt (pow (_x, 2) + pow (_y, 2) );
		}

		// derived vectors
		Vector normal() {
			return (Vector (_y, -_x) );
		}
		Vector singleUnit() {
			return Vector (_x / length(), _y / length() );
		}

		// conversion to 2x uint8
		uint16_t toUint16() {
			uint8_t ix = (int) (_x + 0.5);
			if (_x < 0) ix = (int) (_x - 0.5);
			uint8_t iy = (int) (_y + 0.5);
			if (_y < 0) iy = (int) (_y - 0.5);
			return (iy << 8) + ix;
		}

		// vector algebra
		Vector& operator+= (Vector& other) {
			_x += other.x();
			_y += other.y();
			return *this;
		}
		Vector& operator-= (Vector& other) {
			_x -= other.x();
			_y -= other.y();
			return *this;
		}
		Vector& operator/= (double d) {
			_x /= d;
			_y /= d;
			return *this;
		}
		Vector& operator*= (double d) {
			_x *= d;
			_y *= d;
			return *this;
		}

		Vector operator+ (Vector other) {
			return Vector (_x + other.x(), _y + other.y() );
		}
		Vector operator- (Vector other) {
			return Vector (_x - other.x(), _y - other.y() );
		}
		Vector operator* (double d) {
			return Vector (_x * d, _y * d);
		}
		Vector operator/ (double d) {
			return Vector (_x / d, _y / d);
		}

		Vector intersectBorder (Vector& nv, int x, int y, direction d) {
			x--;
			y--;
			Vector n = nv;
			if (d == RIGHT) n = Vector (-nv.x(), -nv.y() );
			double t1, t2, t3, t4;
			if (n.x() == 0) {
				t1 = 1e6;
				t2 = 1e6;
			} else {
				t1 = -_x / n.x();
				t2 = (x - _x) / n.x();
			}
			if (n.y() == 0) {
				t3 = 1e6;
				t4 = 1e6;
			} else {
				t3 = -_y / n.y();
				t4 = (y - _y) / n.y();
			}
			t1 = (t1 < 0) ? 1e6 : t1;
			t2 = (t2 < 0) ? 1e6 : t2;
			t3 = (t3 < 0) ? 1e6 : t3;
			t4 = (t4 < 0) ? 1e6 : t4;
			double a = (t1 < t2) ? t1 : t2;
			double b = (t3 < t4) ? t3 : t4;
			double min_t = (a < b) ? a : b;
			if (min_t == 1e6) throw ("Fatal error in vector calculation");
			int rx = (int) (_x + n.x() * min_t);
			while (rx < 0) rx ++;
			while (rx > x) rx --;
			int ry = (int) (_y + n.y() * min_t);
			while (ry < 0) ry ++;
			while (ry > y) ry --;
			return Vector (rx, ry);
		}
};

typedef Vector Point;

#endif // VECTOR_H
