/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*- */
/*
 * fische-3.1
 * Copyright (C) Marcel Ebmer 2009-2010 <marcel@26elf.at>
 *
 * fische-3.1 is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * fische-3.1 is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <iostream>
#include <stdint.h>
#include <cstdlib>
#include <cmath>
#include "sdlscreen.h"
#include "vectorfield.h"
#include "ringbuffer.h"
#include "wavepainter.h"
#include "configuration.h"

using namespace std;

WavePainter* WavePainter::ms_instance = 0;

WavePainter::WavePainter () {
	VectorField* v = VectorField::instance();
	SdlScreen* s = SdlScreen::instance();

	__height = v->height();
	__width = v->width();
	__y0 = (s->height() - v->height() ) / 2;
	__x0 = (s->width() - v->width() ) / 2;
	__ringbuffer = RingBuffer::instance();
	__sdlscreen = s;
	__xCenter = __x0 + __width / 2;
	__yCenter = __y0 + __height / 2;
	__shape = 1;
	__shapes = 2;
	__color1 = rand() % 0xffffffff;
	__color2 = ~__color1;
	__alpha = 0;
	__dir = 1;
	__state = STEADY;
	__inc = 0;
}

WavePainter* WavePainter::instance () {
	if (!ms_instance) ms_instance = new WavePainter();
	return ms_instance;
}

void WavePainter::line (int x1, int y1, int x2, int y2, uint32_t color) {
	int Dx = abs (x1 - x2);
	int Dy = abs (y1 - y2);
	int dx = 1;
	if (x2 < x1) dx = -1;
	int dy = 1;
	if (y2 < y1) dy = -1;
	uint32_t* pixels = (uint32_t*) __sdlscreen->pixels();
	int pitch = __sdlscreen->width();

	if ( (Dx == 0) && (Dy == 0) ) return;

	if (Dx > Dy) {
		for (int x = x1; x * dx <= x2 * dx; x += dx) {
			int y = (int) ( (double) y1 + (double) Dy / (double) Dx * (double) dy * (double) abs (x - x1) + 0.5);
			if ( (x < __x0) || (x >= __x0 + __width) || (y < __y0) || (y >= __y0 + __height) ) continue;
			* (pixels + y * pitch + x) = color;
		}
	} else {
		for (int y = y1; y * dy <= y2 * dy; y += dy) {
			int x = (int) ( (double) x1 + (double) Dx / (double) Dy * (double) dx * (double) abs (y - y1) + 0.5);
			if ( (x < __x0) || (x >= __x0 + __width) || (y < __y0) || (y >= __y0 + __height) ) continue;
			* (pixels + y * pitch + x) = color;
		}
	}
}

void WavePainter::paint () {
	const double Pi = acos (-1);
	static double f = 0;
	switch (__state) {
		case STEADY:		// waiting for valid bpm
			break;
		case ROTATING:
			__alpha += __inc;
			if ( (__alpha > 2 * Pi) || (__alpha < -2 * Pi) ) {
				__alpha = 0;
				__state = STEADY;
			}
	}

	int16_t* samples = (int16_t*) __ringbuffer->get();
	int l = __ringbuffer->size() / 4;

	int dim = (__height < __width) ? __height : __width;
	int div = 196605 / dim;

	switch (__shape) {
		case 0: {
			Point base1 (__width / 2 + dim / 6 * sin (__alpha), __height / 2 + dim / 6 * cos (__alpha) );
			Vector n = base1 - Vector (__width / 2, __height / 2);
			n = n.normal();
			Point start1 = base1.intersectBorder (n, __width, __height, Vector::LEFT);
			Point end1 = base1.intersectBorder (n, __width, __height, Vector::RIGHT);

			Point base2 (__width / 2 - dim / 6 * sin (__alpha), __height / 2 - dim / 6 * cos (__alpha) );
			n = base2 - Vector (__width / 2, __height / 2);
			n = n.normal();
			Point start2 = base2.intersectBorder (n, __width, __height, Vector::LEFT);
			Point end2 = base2.intersectBorder (n, __width, __height, Vector::RIGHT);

			Vector v1 = end1 - start1;
			v1 /= l;
			Vector v2 = end2 - start2;
			v2 /= l;
			Vector n1 = v1.singleUnit().normal();
			Vector n2 = v2.singleUnit().normal();

			Point base, pt1, pt2;
			for (int i = 0; i < l - 1; i ++) {
				base = start1 + v1 * i;
				pt1 = base + n1 * (* (samples + 2 * i) / div);
				base += v1;
				pt2 = base + n1 * (* (samples + 2 * (i + 1) ) / div);
				line (pt1, pt2, __color1);
				base = start2 + v2 * i;
				pt1 = base + n2 * (* (samples + 1 + 2 * i) / div);
				base += v2;
				pt2 = base + n2 * (* (samples + 1 + 2 * (i + 1) ) / div);
				line (pt1, pt2, __color2);
			}
			return;
		}
		case 1:
			f = cos (Pi / 3 + 2 * __alpha) + 0.5;
			for (int i = 0; i < l - 1; i ++) {
				double phi1 = Pi * (0.25 + (double) i / (double) l) + __alpha;
				double phi2 = phi1 + Pi / (double) l;
				double r1 = dim / 4 + * (samples + 2 * i) / div;
				double r2 = dim / 4 + * (samples + 2 * (i + 1) ) / div;
				int x1 = iround (__xCenter + f * r1 * sin (phi1) );
				int x2 = iround (__xCenter + f * r2 * sin (phi2) );
				int y1 = iround (__yCenter + r1 * cos (phi1) );
				int y2 = iround (__yCenter + r2 * cos (phi2) );
				line (x1, y1, x2, y2, __color1);
				phi1 += Pi;
				phi2 += Pi;
				r1 = dim / 4 + * (samples + 1 + 2 * i) / div;
				r2 = dim / 4 + * (samples + 1 + 2 * (i + 1) ) / div;
				x1 = iround (__xCenter + f * r1 * sin (phi1) );
				x2 = iround (__xCenter + f * r2 * sin (phi2) );
				y1 = iround (__yCenter + r1 * cos (phi1) );
				y2 = iround (__yCenter + r2 * cos (phi2) );
				line (x1, y1, x2, y2, __color2);
			}
			return;
	}
}

void WavePainter::changeShape() {
	if (__state == ROTATING) return;
	int n = rand() % __shapes;
	while (n == __shape) n = rand() % __shapes;
	__shape = n;
}

void WavePainter::beat (double bpm) {
	const double Pi = acos (-1);
	switch (__state) {
		case STEADY:		// waiting for valid bpm
			if (!bpm) break;
		//	if (rand() % 3 != 0) break;
			else {
				__dir = 1 - 2 * (rand() % 2);
				__inc = Pi / bpm / 2 * __dir;
				__alpha = 0;
				__state = ROTATING;
			}
		case ROTATING:
			break;
	}
}

void WavePainter::changeColor() {
	__color1 = rand() % 0xffffffff;
	__color2 = ~__color1;
}

int iround (double v) {
	if (v > 0) return (int) (v + 0.5);
	else return (int) (v - 0.5);
}
